package cs108;

import java.util.Iterator;
import java.util.NoSuchElementException;

public final class SArrayList<E> extends SAbstractList<E> {
    private int size = 0;
    @SuppressWarnings("unchecked")
    private E[] array = (E[]) new Object[10];

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(int i, E e) {
        checkPositionIndex(i);
        if (size < array.length) {
            System.arraycopy(array, i, array, i + 1, size - i);
        } else {
            @SuppressWarnings("unchecked")
            E[] newArray = (E[]) new Object[array.length * 2];
            System.arraycopy(array, 0, newArray, 0, i);
            System.arraycopy(array, i, newArray, i + 1, size - i);
            array = newArray;
        }
        array[i] = e;
        size += 1;
    }

    @Override
    public void remove(int i) {
        checkElementIndex(i);
        System.arraycopy(array, i + 1, array, i, size - i - 1);
        array[--size] = null;
    }

    @Override
    public E get(int i) {
        return array[checkElementIndex(i)];
    }

    @Override
    public E set(int i, E e) {
        E oldE = array[checkElementIndex(i)];
        array[i] = e;
        return oldE;
    }

    // Itérateur sous forme de classe intérieure anonyme.
    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private int nextI = 0;
            private boolean canRemove = false;

            @Override
            public boolean hasNext() {
                return nextI < size;
            }

            @Override
            public E next() {
                if (!hasNext())
                    throw new NoSuchElementException();
                canRemove = true;
                return array[nextI++];
            }

            @Override
            public void remove() {
                if (!canRemove)
                    throw new IllegalStateException();
                canRemove = false;
                SArrayList.this.remove(--nextI);
            }
        };
    }

    // Autres solutions (moins bonnes) pour l'itérateur:

    // 1. classe imbriquée statiquement
    // nécessite une autre mise en œuvre de la méthode iterator:
//    @Override public Iterator<E> iterator() {
//        return new SALIterator1<>(this);
//    }

    private static final class SALIterator1<E> implements Iterator<E> {
        private final SArrayList<E> list;
        private int nextI;
        private boolean canRemove;

        public SALIterator1(SArrayList<E> list) {
            this.list = list;
            this.nextI = 0;
            this.canRemove = false;
        }

        @Override
        public boolean hasNext() {
            return nextI < list.size;
        }

        @Override
        public E next() {
            if (! hasNext())
                throw new NoSuchElementException();
            canRemove = true;
            return list.array[nextI++];
        }

        @Override
        public void remove() {
            if (! canRemove)
                throw new IllegalStateException();
            canRemove = false;
            list.remove(--nextI);
        }
    }

    // 2. classe intérieure (c-à-d imbriquée non statiquement)
    // nécessite une autre mise en œuvre de la méthode iterator:
//    @Override public Iterator<E> iterator() {
//        return new SALIterator2(); // ou encore : return this.new SALIterator2();
//    }

    private final class SALIterator2 implements Iterator<E> {
        private int nextI;
        private boolean canRemove;

        @Override
        public boolean hasNext() {
            return nextI < size;
        }

        @Override
        public E next() {
            if (! hasNext())
                throw new NoSuchElementException();
            canRemove = true;
            return array[nextI++];
        }

        @Override
        public void remove() {
            if (! canRemove)
                throw new IllegalStateException();
            canRemove = false;
            SArrayList.this.remove(--nextI);
        }
    }
}
