package cs108;

import java.util.Iterator;
import java.util.NoSuchElementException;

public final class SLinkedList<E> extends SAbstractList<E> {
    private int size = 0;
    private Node<E> head = null;

    @Override
    public int size() {
        return size;
    }

    @Override
    public void add(int i, E e) {
        if (checkPositionIndex(i) == 0) {
            head = new Node<>(head, e);
        } else {
            Node<E> pred = getNode(i - 1);
            pred.next = new Node<>(pred.next, e);
        }
        size += 1;
    }

    @Override
    public void remove(int i) {
        if (checkElementIndex(i) == 0) {
            head = head.next;
        } else {
            Node<E> pred = getNode(i - 1);
            pred.next = pred.next.next;
        }
        size -= 1;
    }

    @Override
    public E get(int i) {
        return getNode(checkElementIndex(i)).elem;
    }

    @Override
    public E set(int i, E e) {
        Node<E> node = getNode(checkElementIndex(i));
        E oldE = node.elem;
        node.elem = e;
        return oldE;
    }

    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private Node<E> pred = null, curr = null, next = head;
            private boolean canRemove = false;

            @Override
            public boolean hasNext() {
                return next != null;
            }

            @Override
            public E next() {
                if (!hasNext())
                    throw new NoSuchElementException();
                canRemove = true;

                E elem = next.elem;
                pred = curr;
                curr = next;
                next = next.next;
                return elem;
            }

            @Override
            public void remove() {
                if (!canRemove)
                    throw new IllegalStateException();
                canRemove = false;

                if (pred == null)
                    head = head.next;
                else
                    pred.next = pred.next.next;
                curr = pred;
                pred = null;
                size -= 1;
            }
        };
    }

    // Itérateur non modifiable (plus simple à comprendre initialement)
    public Iterator<E> unmodifiableIterator() {
        return new Iterator<E>() {
            private Node<E> next = head;

            @Override
            public boolean hasNext() {
                return next != null;
            }

            @Override
            public E next() {
                if (!hasNext())
                    throw new NoSuchElementException();

                E elem = next.elem;
                next = next.next;
                return elem;
            }
        };
    }

    private Node<E> getNode(int i) {
        Node<E> n = head;
        for (int j = 0; j < i; ++j)
            n = n.next;
        return n;
    }

    private static final class Node<E> {
        private Node<E> next;
        private E elem;

        public Node(Node<E> next, E elem) {
            this.next = next;
            this.elem = elem;
        }
    }
}
