package cs108;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Random;

import org.junit.jupiter.api.Test;

public abstract class SListTest {
    protected abstract <E> SList<E> newList();

    @Test
    public void sizeIsInitially0() {
        assertEquals(0, newList().size());
    }

    @Test
    public void sizeIncreasesWhenAdding() {
        SList<Integer> l = newList();
        for (int i = 0; i < 10; ++i) {
            assertEquals(i, l.size());
            l.add(0, i);
        }
    }

    @Test
    public void sizeDecreasesWhenRemoving() {
        SList<Integer> l = newList();
        for (int i = 0; i < 10; ++i)
            l.add(l.size(), i);
        for (int s = 10; s > 0; --s) {
            assertEquals(s, l.size());
            l.remove(0);
        }
        assertEquals(0, l.size());
    }

    @Test
    public void toStringWorksOnEmptyList() {
        assertEquals("[]", newList().toString());
    }
    
    private enum Action { ADD, REMOVE, SET }

    @Test
    public void behavesLikeJavaList() {
        Random rng = new Random();
        
        List<Integer> jl = new ArrayList<>();
        SList<Integer> l = newList();
        
        for (int i = 1; i < 10000; ++i) {
            Action action = jl.isEmpty()
                    ? Action.ADD
                            : Action.values()[rng.nextInt(Action.values().length)];
            switch (action) {
            case ADD: {
                int e = rng.nextInt();
                int j = rng.nextInt(l.size() + 1);
                jl.add(j, e);
                l.add(j, e);
            } break;
            
            case REMOVE: {
                int j = rng.nextInt(l.size());
                jl.remove(j);
                l.remove(j);
            } break;

            case SET: {
                int e1 = rng.nextInt();
                int j = rng.nextInt(l.size());
                jl.set(j, e1);
                l.set(j, e1);
            } break;
            
            default:
                throw new Error();
            }
            
            assertEquals(jl.size(), l.size());
            assertPairwiseEqual(jl, l);
        }

        Iterator<Integer> ji = jl.iterator();
        Iterator<Integer> i = l.iterator();
        while (ji.hasNext()) {
            assertTrue(i.hasNext());
            int je = ji.next(), e = i.next();
            assertEquals(je, e);
            if (rng.nextBoolean()) {
                ji.remove();
                i.remove();
            }
        }
        assertPairwiseEqual(jl, l);
    }
    
    private static <E> void assertPairwiseEqual(List<E> jl, SList<E> l) {
        Iterator<E> ji = jl.iterator();
        Iterator<E> i = l.iterator();
        while (ji.hasNext())
            assertEquals(ji.next(), i.next());
        assertFalse(i.hasNext());
    }
}
